@@ -244,7 +244,7 @@ def StoreCacheDatum(ticker, objectName, datum, expiry=None, metadata=None):
 	fp = GetFilepath(ticker, objectName, object=datum, prune=True)
 
 	if verbose:
-		print("- storing {} at {}".format(objectName, fp))
+		print("- storing {} as {} at {}".format(objectName, ext, fp))
 
 	d = _ReadData(ticker, objectName)
 	if d is None:
@@ -332,11 +332,7 @@ def ReadCacheMetadata(ticker, objectName, key):
 			data   = objData["data"]
 			md     = objData["metadata"] if "metadata" in objData else None
 	else:
-		d = _ReadData(ticker, objectName)
-		md = d["metadata"]
-		if verbose:
-			print("ReadCacheMetadata() read md as:")
-			print(md)
+		data,md = _ReadData(ticker, objectName, return_metadata_too=True)
 	if md is None:
 		return None
 	elif not key in md:
@@ -358,10 +354,6 @@ def WriteCacheMetadata(ticker, objectName, key, value):
 	else:
 		d["metadata"] = {key:value}
 
-	if verbose:
-		print("WriteCacheMetadata() updated md to:")
-		print(d["metadata"])
-
 	fp = GetFilepath(ticker, objectName)
 	if fp.endswith(".json"):
 		with open(fp, 'w') as outData:
